@@ -6,7 +6,7 @@ from typing import Callable, Dict, List, Tuple, Union
 import numpy as np
 import torch
 
-from neurosym.dsl.dsl import DSL
+from neurosym.dsl.dsl import DSL, ROOT_SYMBOL
 from neurosym.program_dist.tree_distribution.ordering import DefaultNodeOrdering
 from neurosym.program_dist.tree_distribution.preorder_mask.preorder_mask import (
     ConjunctionPreorderMask,
@@ -24,22 +24,77 @@ from .tree_distribution.tree_distribution import TreeProgramDistributionFamily
 
 @dataclass
 class BigramProgramDistribution:
+    dist_fam: "BigramProgramDistributionFamily"
     distribution: np.ndarray
 
     def __post_init__(self):
         assert self.distribution.ndim == 3
         assert self.distribution.shape[0] == self.distribution.shape[2]
 
-    def bound_minimum_likelihood(self, min_likelihood: float):
+    def bound_minimum_likelihood(
+        self, min_likelihood: float, symbol_mask: np.ndarray = None
+    ):
         assert 0 <= min_likelihood <= 1
+        if symbol_mask is not None:
+            assert (
+                len(symbol_mask.shape) == 1
+                and symbol_mask.shape[0] == self.distribution.shape[0]
+            )
+            assert symbol_mask.dtype == np.bool_
         distribution = self.distribution
-        distribution = np.maximum(distribution, min_likelihood)
-        distribution = distribution / distribution.sum(-1)[..., None]
-        return BigramProgramDistribution(distribution)
+        if symbol_mask is None:
+            distribution = np.maximum(distribution, min_likelihood)
+        else:
+            mask_square = self._square_mask(symbol_mask)
+            distribution = distribution.copy()
+            distribution[mask_square] = np.maximum(
+                distribution[mask_square], min_likelihood
+            )
+        distribution = self.dist_fam.mask_invalid(distribution)
+        distribution = distribution / (distribution.sum(-1)[..., None] + 1e-10)
+        return BigramProgramDistribution(self.dist_fam, distribution)
+
+    def _square_mask(self, symbol_mask):
+        mask_square = symbol_mask[:, None] & symbol_mask[None, :]
+        mask_square = mask_square[:, None, :].repeat(self.distribution.shape[1], axis=1)
+
+        return mask_square
+
+    def mix_with_other(self, other: "BigramProgramDistribution", weight_other: float):
+        # pylint: disable=self-cls-assignment
+        assert 0 <= weight_other <= 1
+        symbols_this = self.dist_fam.symbols()
+        symbols_other = other.dist_fam.symbols()
+        if set(symbols_this).issubset(symbols_other):
+            self, other = other, self
+            symbols_this, symbols_other = symbols_other, symbols_this
+            weight_other = 1 - weight_other
+        else:
+            if not set(symbols_this).issuperset(symbols_other):
+                extra_this = set(symbols_this) - set(symbols_other)
+                extra_other = set(symbols_other) - set(symbols_this)
+                extra_this, extra_other = (
+                    ", ".join(repr(x) for x in sorted(extra))
+                    for extra in (extra_this, extra_other)
+                )
+                raise ValueError(
+                    "DSL not compatible, extra symbols in this: "
+                    f"{extra_this}, extra symbols in other: {extra_other}"
+                )
+        mask = np.array([s in symbols_other for s in symbols_this], dtype=np.bool_)
+        distribution = self.distribution.copy()
+        distribution_other = np.zeros_like(distribution)
+        distribution_other[self._square_mask(mask)] = other.distribution.flatten()
+        distribution[mask] = (
+            weight_other * distribution_other[mask]
+            + (1 - weight_other) * distribution[mask]
+        )
+        return BigramProgramDistribution(self.dist_fam, distribution)
 
 
 @dataclass
 class BigramProgramDistributionBatch:
+    dist_fam: "BigramProgramDistributionFamily"
     distribution_batch: np.ndarray
 
     def __post_init__(self):
@@ -50,7 +105,7 @@ class BigramProgramDistributionBatch:
         assert self.distribution_batch.shape[1] == self.distribution_batch.shape[3]
 
     def __getitem__(self, i):
-        return BigramProgramDistribution(self.distribution_batch[i])
+        return BigramProgramDistribution(self.dist_fam, self.distribution_batch[i])
 
     def __len__(self):
         return len(self.distribution_batch)
@@ -79,6 +134,7 @@ class BigramProgramCounts:
 
 @dataclass
 class BigramProgramCountsBatch:
+    dist_fam: "BigramProgramDistributionFamily"
     counts: List[BigramProgramCounts]
 
     def numerators(self, num_symbols, max_arity):
@@ -114,7 +170,9 @@ class BigramProgramCountsBatch:
         # a simple conversion from counts to probabilities, and we do
         # not need to handle the case where the denominator is 0.
 
-        return BigramProgramDistributionBatch(counts_to_probabilities(numerators))
+        return BigramProgramDistributionBatch(
+            self.dist_fam, counts_to_probabilities(numerators)
+        )
 
 
 class BigramProgramDistributionFamily(TreeProgramDistributionFamily):
@@ -127,7 +185,7 @@ class BigramProgramDistributionFamily(TreeProgramDistributionFamily):
             Callable[[DSL, TreeDistribution], PreorderMask]
         ] = (),
         include_type_preorder_mask: bool = True,
-        node_ordering=lambda _: DefaultNodeOrdering(),
+        node_ordering=DefaultNodeOrdering,
     ):
         if valid_root_types is not None:
             dsl = dsl.with_valid_root_types(valid_root_types)
@@ -168,7 +226,7 @@ class BigramProgramDistributionFamily(TreeProgramDistributionFamily):
             parameters.shape[1:] == self.parameters_shape()
         ), f"Expected {self.parameters_shape()}, got {parameters.shape[1:]}"
         parameters = self.normalize_parameters(parameters, logits=False)
-        return BigramProgramDistributionBatch(parameters.detach().cpu().numpy())
+        return BigramProgramDistributionBatch(self, parameters.detach().cpu().numpy())
 
     def count_programs(self, data: List[List[SExpression]]) -> BigramProgramCountsBatch:
         tree_dist = self.tree_distribution_skeleton
@@ -178,7 +236,7 @@ class BigramProgramDistributionFamily(TreeProgramDistributionFamily):
             all_counts.append(
                 BigramProgramCounts(numerators=numerators, denominators=denominators)
             )
-        return BigramProgramCountsBatch(all_counts)
+        return BigramProgramCountsBatch(self, all_counts)
 
     def counts_to_distribution(
         self, counts: BigramProgramCountsBatch
@@ -281,7 +339,9 @@ class BigramProgramDistributionFamily(TreeProgramDistributionFamily):
         return -(numer - denom)
 
     def uniform(self):
-        return BigramProgramDistribution(counts_to_probabilities(self._valid_mask))
+        return BigramProgramDistribution(
+            self, counts_to_probabilities(self._valid_mask)
+        )
 
     def compute_tree_distribution(
         self, distribution: Union[BigramProgramDistribution, NoneType]
@@ -290,17 +350,17 @@ class BigramProgramDistributionFamily(TreeProgramDistributionFamily):
             assert isinstance(distribution, BigramProgramDistribution), type(
                 distribution
             )
-            dist = defaultdict(list)
-            for parent, position, child in zip(
-                *np.where(distribution.distribution > 0)
-            ):
-                dist[(parent, position),].append(
-                    (child, np.log(distribution.distribution[parent, position, child]))
-                )
-            dist = {k: sorted(v, key=lambda x: -x[1]) for k, v in dist.items()}
+            dist_vals = distribution.distribution
         else:
             assert distribution is None
-            dist = None
+            dist_vals = self._valid_mask
+
+        dist = defaultdict(list)
+        for parent, position, child in zip(*np.where(dist_vals > 0)):
+            dist[(parent, position),].append(
+                (child, np.log(dist_vals[parent, position, child]))
+            )
+        dist = {k: sorted(v, key=lambda x: -x[1]) for k, v in dist.items()}
 
         return TreeDistribution(
             1,
@@ -316,11 +376,17 @@ class BigramProgramDistributionFamily(TreeProgramDistributionFamily):
             masks.append(TypePreorderMask(tree_dist, self._dsl))
         for mask in self._additional_preorder_masks:
             masks.append(mask(tree_dist, self._dsl))
-        return ConjunctionPreorderMask(tree_dist, masks)
+        return ConjunctionPreorderMask.of(tree_dist, masks)
+
+    def mask_invalid(self, distribution):
+        return distribution * self._valid_mask
+
+    def symbols(self):
+        return self._symbols
 
 
 def bigram_mask(dsl):
-    symbols = ["<root>"] + sorted([x.symbol() for x in dsl.productions])
+    symbols = dsl.ordered_symbols(include_root=True)
 
     valid_root_types = dsl.valid_root_types
     symbol_to_idx = {sym: i for i, sym in enumerate(symbols)}
@@ -341,7 +407,7 @@ def bigram_mask(dsl):
     valid_mask = np.zeros((len(symbols), max(arities), len(symbols)), dtype=np.bool_)
     root_syms = {sym for t in valid_root_types for sym, _ in rules_for[t]}
     for root_sym in root_syms:
-        valid_mask[symbol_to_idx["<root>"], 0, symbol_to_idx[root_sym]] = 1
+        valid_mask[symbol_to_idx[ROOT_SYMBOL], 0, symbol_to_idx[root_sym]] = 1
     for _, rules in rules_for.items():
         for root_sym, child_types in rules:
             for i, ct in enumerate(child_types):
